@@ -36,9 +36,10 @@
 
 static int displayPrivateIndex;
 
-#define SHOT_DISPLAY_OPTION_INITIATE 0
-#define SHOT_DISPLAY_OPTION_DIR      1
-#define SHOT_DISPLAY_OPTION_NUM	     2
+#define SHOT_DISPLAY_OPTION_INITIATE   0
+#define SHOT_DISPLAY_OPTION_DIR        1
+#define SHOT_DISPLAY_OPTION_LAUNCH_APP 2
+#define SHOT_DISPLAY_OPTION_NUM        3
 
 typedef struct _ShotDisplay {
     int		    screenPrivateIndex;
@@ -104,7 +105,7 @@ shotInitiate (CompDisplay     *d,
 	ss->grab = TRUE;
     }
 
-    return FALSE;
+    return TRUE;
 }
 
 static Bool
@@ -182,6 +183,7 @@ shotSort (const void *_a,
 static Bool
 shotPaintScreen (CompScreen		 *s,
 		 const ScreenPaintAttrib *sAttrib,
+		 const CompTransform	 *transform,
 		 Region			 region,
 		 int			 output,
 		 unsigned int		 mask)
@@ -191,7 +193,7 @@ shotPaintScreen (CompScreen		 *s,
     SHOT_SCREEN (s);
 
     UNWRAP (ss, s, paintScreen);
-    status = (*s->paintScreen) (s, sAttrib, region, output, mask);
+    status = (*s->paintScreen) (s, sAttrib, transform, region, output, mask);
     WRAP (ss, s, paintScreen, shotPaintScreen);
 
     if (status && ss->grab)
@@ -251,6 +253,7 @@ shotPaintScreen (CompScreen		 *s,
 		    if (n >= 0)
 		    {
 			char name[256];
+			char *app;
 			int  number = 0;
 
 			if (n > 0)
@@ -265,12 +268,30 @@ shotPaintScreen (CompScreen		 *s,
 
 			sprintf (name, "screenshot%d.png", number);
 
+			app = sd->opt[SHOT_DISPLAY_OPTION_LAUNCH_APP].value.s;
+
 			if (!writeImageToFile (s->display, dir, name, "png",
 					       w, h, buffer))
 			{
 			    fprintf (stderr, "%s: failed to write "
 				     "screenshot image", programName);
 			}
+			else if (*app != '\0')
+			{
+			    char *command;
+
+			    command = malloc (strlen (app) +
+					      strlen (dir) +
+					      strlen (name) + 3);
+			    if (command)
+			    {
+				sprintf (command, "%s %s/%s", app, dir, name);
+
+				runCommand (s, command);
+
+				free (command);
+			    }
+			}
 		    }
 		    else
 		    {
@@ -380,6 +401,15 @@ shotDisplayInitOptions (ShotDisplay *sd)
     o->value.s	      = strdup (SHOT_DIR_DEFAULT);
     o->rest.s.string  = 0;
     o->rest.s.nString = 0;
+
+    o = &sd->opt[SHOT_DISPLAY_OPTION_LAUNCH_APP];
+    o->name           = "launch_app";
+    o->shortDesc      = N_("Launch Application");
+    o->longDesc       = N_("Automatically open screenshot in this application");
+    o->type           = CompOptionTypeString;
+    o->value.s        = strdup ("");
+    o->rest.s.string  = NULL;
+    o->rest.s.nString = 0;
 }
 
 static CompOption *
@@ -414,6 +444,10 @@ shotSetDisplayOption (CompDisplay    *display,
     case SHOT_DISPLAY_OPTION_DIR:
 	if (compSetStringOption (o, value))
 	    return TRUE;
+	break;
+    case SHOT_DISPLAY_OPTION_LAUNCH_APP:
+	if (compSetStringOption (o, value))
+	    return TRUE;
     default:
 	break;
     }
@@ -489,6 +523,10 @@ shotFiniScreen (CompPlugin *p,
 		CompScreen *s)
 {
     SHOT_SCREEN (s);
+    SHOT_DISPLAY (s->display);
+
+    removeScreenAction (s, 
+			&sd->opt[SHOT_DISPLAY_OPTION_INITIATE].value.action);
 
     UNWRAP (ss, s, paintScreen);
 
