@@ -1786,6 +1786,7 @@ waterFiniScreen (CompPlugin *p,
     int		  i;
 
     WATER_SCREEN (s);
+    WATER_DISPLAY (s->display);
 
     if (ws->rainHandle)
 	compRemoveTimeout (ws->rainHandle);
@@ -1818,6 +1819,19 @@ waterFiniScreen (CompPlugin *p,
 	function = next;
     }
 
+    removeScreenAction (s, 
+			&wd->opt[WATER_DISPLAY_OPTION_INITIATE].value.action);
+    removeScreenAction (s,
+	   		&wd->opt[WATER_DISPLAY_OPTION_TOGGLE_RAIN].value.action);
+    removeScreenAction (s,
+	   		&wd->opt[WATER_DISPLAY_OPTION_TOGGLE_WIPER].value.action);
+    removeScreenAction (s, 
+			&wd->opt[WATER_DISPLAY_OPTION_TITLE_WAVE].value.action);
+    removeScreenAction (s, 
+			&wd->opt[WATER_DISPLAY_OPTION_POINT].value.action);
+    removeScreenAction (s, 
+			&wd->opt[WATER_DISPLAY_OPTION_LINE].value.action);
+
     UNWRAP (ws, s, preparePaintScreen);
     UNWRAP (ws, s, donePaintScreen);
     UNWRAP (ws, s, drawWindowTexture);
@@ -1850,7 +1864,8 @@ waterGetVersion (CompPlugin *plugin,
 }
 
 CompPluginDep waterDeps[] = {
-    { CompPluginRuleBefore, "blur" }
+    { CompPluginRuleBefore, "blur" },
+    { CompPluginRuleBefore, "video" }
 };
 
 static CompPluginVTable waterVTable = {
